# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 04.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Classes used to classify texture with GLCM analysis.
"""

# pylint: disable=bad-indentation
# pylint: disable=attribute-defined-outside-init

import sys
import time

import numpy as np
import cv2
from skimage.filters.rank import entropy # pylint: disable=import-error
from scipy import signal # pylint: disable=import-error

def first_order_variance(image, window_order):
  """Compute the local, first order variance of an image."""
  k_height = 2*window_order + 1
  k_width = 2*window_order + 1
  image = image.astype('float64')

  # Var(x) = E[x^2] - (E[x])^2. Therefore, we compute the mean of the square,
  # and the mean, and return their difference.
  mean = cv2.boxFilter(image, -1, (k_height, k_width), normalize=True,
                       borderType=cv2.BORDER_REFLECT)
  mean_sqr = cv2.boxFilter(image*image, -1, (k_height, k_width), normalize=True,
                           borderType=cv2.BORDER_REFLECT)

  return mean_sqr - mean*mean

def first_order_entropy(image, window_order):
  """Compute the local, first order entropy of an image."""

  # The entrpy function uses a mask of 0 and 1 to describe the window, from
  # where the entrpy is computed.
  neighbourhood = np.ones((2*window_order + 1, 2*window_order + 1))

  return entropy(image, neighbourhood)

def laws_energy_features(image):
  """Compute laws energy features"""

  mean_w_size = (10, 10)
  energy_w_size = (35, 35)

  # Subtract mean of image
  image = image.astype('float64')
  mean_image = cv2.boxFilter(image, -1, mean_w_size, normalize=True,
                             borderType=cv2.BORDER_REFLECT)
  pp_image = image - mean_image

  # Linear filters
  filters = {}
  filters['L5'] = np.array([1, 4, 6, 4, 1])
  filters['E5'] = np.array([-1, -2, 0, 2, 1])
  filters['S5'] = np.array([-1, 0, 2, 0, -1])
  filters['R5'] = np.array([1, -4, 6, -4, 1])

  # Use the assosiativity of convolution to produce 16 different filter images.
  # signal.sepfir2d convolves the input with a separable filter defined by 2 1d
  # arrays. Also compute the absolute value, since this is what is used to
  # compute the energy maps.
  filter_images = {}
  for name1, filt1 in filters.items():
    for name2, filt2 in filters.items():
      filter_images[name1 + "_" + name2] = np.abs(signal.sepfir2d(pp_image, filt1, filt2))

  energy_images = {}
  for name, filter_image in filter_images.items():
    energy_images[name] = cv2.boxFilter(filter_image, -1, energy_w_size,
                                        borderType=cv2.BORDER_REFLECT)

  # Finally, compute the last 9 feature images (for some reason 'L5_L5' is
  # excluded)
  feature_images = {}
  feature_images['E5_E5'] = energy_images['E5_E5']
  feature_images['S5_S5'] = energy_images['S5_S5']
  feature_images['R5_R5'] = energy_images['R5_R5']

  # In the exercise it states that you whould average the energy over windows.
  # I am not sure what is meant by that, so I just did regular averages.
  feature_images['L5_E5'] = (energy_images['L5_E5'] + energy_images['E5_L5']) / 2
  feature_images['L5_S5'] = (energy_images['L5_S5'] + energy_images['S5_L5']) / 2
  feature_images['L5_R5'] = (energy_images['L5_R5'] + energy_images['R5_L5']) / 2
  feature_images['E5_S5'] = (energy_images['E5_S5'] + energy_images['S5_E5']) / 2
  feature_images['E5_R5'] = (energy_images['E5_R5'] + energy_images['R5_E5']) / 2
  feature_images['S5_R5'] = (energy_images['S5_R5'] + energy_images['R5_S5']) / 2

  return feature_images

class GLCMFeatureImage(object):
  """Local class containing a feature image"""
  def __init__(self, image, name, postfix):
    self.image = image
    self.name = name
    self.computation_time = 0.0
    self.postfix = postfix
    self.plot_name = name + '_' + postfix

  def update_image(self, row_ind, col_ind, value):
    """Update feature image values"""
    self.image[row_ind, col_ind] = value

  def scale_image(self, min_val, max_val):
    """Scale feature image"""
    im_max = np.max(self.image)
    im_min = np.min(self.image)
    self.image = min_val + (1 - (im_max - self.image)/(im_max - im_min))*(max_val - min_val)

  def update_time(self, added_time):
    """Update computation time"""
    self.computation_time += added_time

class Texture(object):
  """Texture class"""
  def __init__(self, image, window_order=1, delta_x=1, delta_y=0,
               isotropic_glcm=False, boundary_treatment='mirrored',
               feature_list=None):

    # Extract parameters
    self.original_image = image
    self.num_image_rows, self.num_image_cols = image.shape

    self.greylevels = np.unique(image)
    self.num_greylevels = len(self.greylevels)

    self.window_order = window_order
    self.delta_x = delta_x
    self.delta_y = delta_y
    self.boundary_treatment = boundary_treatment

    self.isotropic_glcm = isotropic_glcm

    if feature_list is None:
      feature_list = ['max_prob',
                      'entropy',
                      'uniformity',
                      'homogeneity',
                      'glcm_var',
                      'inertia',  # Contrast, sum of square variance
                      'correlation',
                      'cluster_shade'
                      'energy_fraction_q1'
                      'energy_fraction_q2'
                      'energy_fraction_q3'
                      'energy_fraction_q4'
                     ]

    self.feature_list = feature_list
    self.feature_images = {}  # Storage of GLCMFeatureImage objects


  def compute_feature(self, name, row_ind, col_ind, glcm, flat_glcm):
    """Implementation of a variety of glcm features identified by their name."""
    if name == 'max_prob':
      # Maximum probability
      temp_time = time.time()
      max_prob = np.max(flat_glcm)
      self.feature_images['max_prob'].update_image(row_ind, col_ind, max_prob)
      self.feature_images['max_prob'].update_time(time.time() - temp_time)

    elif name == 'entropy':
      # Entropy
      temp_time = time.time()
      entr = -np.dot(flat_glcm, np.ma.log(flat_glcm))
      self.feature_images['entropy'].update_image(row_ind, col_ind, entr)
      self.feature_images['entropy'].update_time(time.time() - temp_time)

    elif name == 'uniformity':
      # Uniformity
      temp_time = time.time()
      uniformity = np.dot(flat_glcm, flat_glcm)
      self.feature_images['uniformity'].update_image(row_ind, col_ind, uniformity)
      self.feature_images['uniformity'].update_time(time.time() - temp_time)

    elif name == 'homogeneity':
      # Homogeneity
      temp_time = time.time()
      homogeneity = np.dot(self.hom_weights, flat_glcm)
      self.feature_images['homogeneity'].update_image(row_ind, col_ind, homogeneity)
      self.feature_images['homogeneity'].update_time(time.time() - temp_time)

    elif name == 'glcm_var':
      # GLCM variance
      temp_time = time.time()
      var_weights = np.tile((self.greylevels - np.mean(flat_glcm))*
                            (self.greylevels - np.mean(flat_glcm)),
                            self.num_greylevels)
      glcm_var = np.dot(var_weights, flat_glcm)
      self.feature_images['glcm_var'].update_image(row_ind, col_ind, glcm_var)
      self.feature_images['glcm_var'].update_time(time.time() - temp_time)

    elif name == 'inertia':
      # Inertia
      temp_time = time.time()
      inertia = np.dot(self.con_weights, flat_glcm)
      self.feature_images['inertia'].update_image(row_ind, col_ind, inertia)
      self.feature_images['inertia'].update_time(time.time() - temp_time)

    elif name == 'correlation':
      # Correlation
      temp_time = time.time()
      p_x = np.sum(glcm, axis=0)
      p_y = np.sum(glcm, axis=1)
      mu_x = np.dot(self.greylevels, p_x)
      mu_y = np.dot(self.greylevels, p_y)
      d_x = (self.greylevels - mu_x)**2
      d_y = (self.greylevels - mu_y)**2
      sigma_x = np.sqrt(np.dot(self.greylevels, d_x))
      sigma_y = np.sqrt(np.dot(self.greylevels, d_y))
      correlation = ((np.dot(self.greylevel_prod, flat_glcm) - mu_x*mu_y) /
                     (sigma_x*sigma_y))
      self.feature_images['correlation'].update_image(row_ind, col_ind, correlation)
      self.feature_images['correlation'].update_time(time.time() - temp_time)

    elif name == 'cluster_shade':
      # Cluster shade
      temp_time = time.time()
      p_x = np.sum(glcm, axis=0)
      p_y = np.sum(glcm, axis=1)
      mu_x = np.dot(self.greylevels, p_x)
      mu_y = np.dot(self.greylevels, p_y)
      temp_clu = (self.clu_weights - mu_x - mu_y)**3
      cluster_shade = np.dot(temp_clu, flat_glcm)
      self.feature_images['cluster_shade'].update_image(row_ind,
                                                        col_ind,
                                                        cluster_shade)
      self.feature_images['cluster_shade'].update_time(time.time() - temp_time)

    elif name == 'energy_fraction_q1':
      # Q1 energy fraction
      temp_time = time.time()
      M, N = glcm.shape
      numerator = np.sum(glcm[:M/2, :N/2])
      denominator = np.sum(glcm)
      self.feature_images['energy_fraction_q1'].update_image(
          row_ind,
          col_ind,
          float(numerator)/denominator)
      self.feature_images['energy_fraction_q1'].update_time(time.time() - temp_time)

    elif name == 'energy_fraction_q2':
      # Q2 energy fraction
      temp_time = time.time()
      M, N = glcm.shape
      numerator = np.sum(glcm[:M/2, N/2:])
      denominator = np.sum(glcm)
      self.feature_images['energy_fraction_q2'].update_image(
          row_ind,
          col_ind,
          float(numerator)/denominator)
      self.feature_images['energy_fraction_q2'].update_time(time.time() - temp_time)

    elif name == 'energy_fraction_q3':
      # Q3 energy fraction
      temp_time = time.time()
      M, N = glcm.shape
      numerator = np.sum(glcm[M/2:, :N/2])
      denominator = np.sum(glcm)
      self.feature_images['energy_fraction_q3'].update_image(
          row_ind,
          col_ind,
          float(numerator)/denominator)
      self.feature_images['energy_fraction_q3'].update_time(time.time() - temp_time)

    elif name == 'energy_fraction_q4':
      # Q4 energy fraction
      temp_time = time.time()
      M, N = glcm.shape
      numerator = np.sum(glcm[M/2:, N/2:])
      denominator = np.sum(glcm)
      self.feature_images['energy_fraction_q4'].update_image(
          row_ind,
          col_ind,
          float(numerator)/denominator)
      self.feature_images['energy_fraction_q4'].update_time(time.time() - temp_time)
    else:
      print(name, ' is not yet implemented as a feature')


  @staticmethod
  def compute_glcm(image, delta_x, delta_y, levels,
                   symmetric=None, normed=None, isotropic=None,
                   weights=None):
    """Computes and returns a co-occurance matrix of the input image.

    delta_x and delta_y determines the location of the accompanying cell
    relative to the original cell.

    levels is here a sorted list of unique greylevels.
    All greylevels in image should be present in the levels list, but the levels
    list can contain greylevels not present in image.
    """
    if symmetric is None:
      symmetric = False
    if normed is None:
      normed = False
    if isotropic is None:
      isotropic = False
    if weights is not None:
      weights = list(np.array(weights) / float(sum(weights)))
      isotropic = True
    if (isotropic) and (weights is None):
      weights = [0.25, 0.25, 0.25, 0.25]

    M, N = image.shape
    delta_i = -delta_y
    delta_j = delta_x
    levels = list(levels)

    # Traverse all pairs (i,j) and (i+delta_i,j+delta_j) and increment the
    # co-occurance matrix in (I(g1), I(g2)), for a graylevel shift between the two cells
    # from g1 in (i,j) to g2 in (i+delta_i, j+delta_j). Here, the operator I is a
    # map between a greylevel and an index in [0, num_greylevels-1]. This map is
    # determined by levels such that if levels = [g0, g1, ...] then I(g1) = 1 etc.

    num_levels = len(levels)
    cooccurance_matrix = np.zeros([num_levels, num_levels])

    if isotropic:
      dx_list = [1, 1, 0, -1]
      dy_list = [0, 1, 1, 1]

      dx_list = [i*abs(delta_x) for i in dx_list]
      dy_list = [i*abs(delta_x) for i in dy_list]

      for k in range(4):
        delta_i = -dy_list[k]
        delta_j = dx_list[k]
        weight = weights[k]

        if (delta_i >= 0) and (delta_j >= 0):
          i_range = list(range(M - delta_i))
          j_range = list(range(N - delta_j))
        if (delta_i >= 0) and (delta_j < 0):
          i_range = list(range(M - delta_i))
          j_range = list(range(-delta_j, N))
        if (delta_i < 0) and (delta_j >= 0):
          i_range = list(range(-delta_i, M))
          j_range = list(range(N - delta_j))
        if (delta_i < 0) and (delta_j < 0):
          i_range = list(range(-delta_i, M))
          j_range = list(range(-delta_j, N))

        for i in i_range:
          old_i = i
          new_i = i + delta_i
          for j in j_range:
            old_j = j
            new_j = j + delta_j
            gl_old = image[old_i, old_j]
            gl_new = image[new_i, new_j]
            co_i = levels.index(gl_old)
            co_j = levels.index(gl_new)
            cooccurance_matrix[co_i, co_j] += weight
    else:
      if (delta_i >= 0) and (delta_j >= 0):
        i_range = list(range(M - delta_i))
        j_range = list(range(N - delta_j))
      if (delta_i >= 0) and (delta_j < 0):
        i_range = list(range(M - delta_i))
        j_range = list(range(-delta_j, N))
      if (delta_i < 0) and (delta_j >= 0):
        i_range = list(range(-delta_i, M))
        j_range = list(range(N - delta_j))
      if (delta_i < 0) and (delta_j < 0):
        i_range = list(range(-delta_i, M))
        j_range = list(range(-delta_j, N))

      for i in i_range:
        old_i = i
        new_i = i + delta_i
        for j in j_range:
          old_j = j
          new_j = j + delta_j
          gl_old = image[old_i, old_j]
          gl_new = image[new_i, new_j]
          co_i = levels.index(gl_old)
          co_j = levels.index(gl_new)
          cooccurance_matrix[co_i, co_j] += 1

    if symmetric:
      cooccurance_matrix += cooccurance_matrix.T
      if normed:
        return cooccurance_matrix / float(sum(sum(cooccurance_matrix)))
      else:
        return cooccurance_matrix
    else:
      if normed:
        return cooccurance_matrix / float(sum(sum(cooccurance_matrix)))
      else:
        return cooccurance_matrix

  @staticmethod
  def square_image_window(image, neigh_order, center_row, center_col, boundary):
    """Given an input image, neighbourhood order and center index, it creates a
    square window and returns it. This function is horrible, and it should
    really be some easier way.

    Args:
      image: 2D numpy array. Image to crop the image from.
      neigh_order: int. The 'radius' of a square neighbourhood.
      center_row: int. Center row coordinate.
      center_col: int. Center column coordinate.
      boundary: Either int or string. Specifies boundary treatment when the
                window is outside the boundary.
    """
    if not isinstance(boundary, str):
      boundary_type = 'constant'
      boundary_val = boundary
    elif boundary is 'mirrored':
      boundary_type = 'mirrored'
    elif boundary is 'clamped':
      boundary_type = 'clamped'
    elif boundary is 'warped':
      boundary_type = 'warped'
    else:
      print('{0} boundary type is not implemented yet'.format(boundary))
      print('Constant boundary 0 is chosen')
      boundary_type = 'constant'
      boundary_val = 0

    # Image and window boundaries (inclusive intervals)
    image_min_row = 0
    image_min_col = 0
    image_max_row = image.shape[0] - 1
    image_max_col = image.shape[1] - 1
    wind_glob_min_row = center_row - neigh_order
    wind_glob_min_col = center_col - neigh_order
    wind_glob_max_row = center_row + neigh_order
    wind_glob_max_col = center_col + neigh_order
    wind_loc_max_row = 2*neigh_order
    wind_loc_max_col = 2*neigh_order
    wind_num_rows = wind_loc_max_row + 1
    wind_num_cols = wind_loc_max_col + 1

    # Dimension checks
    if ((center_row < image_min_row) or
            (center_row > image_max_row) or
            (center_col < image_min_col) or
            (center_col > image_max_col)):
      print('Unvalid center index in window')
      print('Provided center index: ({0},{1})'.format(center_row, center_col))
      print('image rows: [{0},{1}]'.format(image_min_row, image_max_row))
      print('image cols: [{0},{1}]'.format(image_min_col, image_max_col))
      sys.exit(1)
    if (wind_num_rows > image.shape[0]) or (wind_num_cols > image.shape[1]):
      print('Size of window is larger than original image')
      sys.exit(1)

    window_image = np.zeros([wind_num_rows, wind_num_cols])

    if boundary_type is 'constant':
      loc_row = 0
      for glob_row in range(wind_glob_min_row, wind_glob_max_row+1):
        loc_col = 0
        if (glob_row < image_min_row) or (glob_row > image_max_row):
          window_image[loc_row, :] = boundary_val
          loc_row += 1
          continue
        for glob_col in range(wind_glob_min_col, wind_glob_max_col+1):
          if (glob_col < image_min_col) or (glob_col > image_max_col):
            window_image[loc_row, loc_col] = boundary_val
            loc_col += 1
            continue
          window_image[loc_row, loc_col] = image[glob_row, glob_col]
          loc_col += 1
        loc_row += 1

    elif boundary_type is 'mirrored':
      val_row = 0
      val_col = 0
      loc_row = 0
      for glob_row in range(wind_glob_min_row, wind_glob_max_row+1):
        loc_col = 0
        for glob_col in range(wind_glob_min_col, wind_glob_max_col+1):
          if glob_row < image_min_row:
            val_row = 2*image_min_row - glob_row - 1
          elif glob_row > image_max_row:
            val_row = image_max_row - np.mod(glob_row, image_max_row)
          else:
            val_row = glob_row

          if glob_col < image_min_col:
            val_col = 2*image_min_col - glob_col - 1
          elif glob_col > image_max_col:
            val_col = image_max_col - np.mod(glob_col, image_max_col)
          else:
            val_col = glob_col

          window_image[loc_row, loc_col] = image[val_row, val_col]
          loc_col += 1
        loc_row += 1

    elif boundary_type is 'clamped':
      val_row = 0
      val_col = 0
      loc_row = 0
      for glob_row in range(wind_glob_min_row, wind_glob_max_row+1):
        loc_col = 0
        for glob_col in range(wind_glob_min_col, wind_glob_max_col+1):
          if glob_row < image_min_row:
            val_row = image_min_row
          elif glob_row > image_max_row:
            val_row = image_max_row
          else:
            val_row = glob_row

          if glob_col < image_min_col:
            val_col = image_min_col
          elif glob_col > image_max_col:
            val_col = image_max_col
          else:
            val_col = glob_col

          window_image[loc_row, loc_col] = image[val_row, val_col]
          loc_col += 1
        loc_row += 1

    elif boundary_type is 'wrapped':
      val_row = 0
      val_col = 0
      loc_row = 0
      for glob_row in range(wind_glob_min_row, wind_glob_max_row+1):
        loc_col = 0
        for glob_col in range(wind_glob_min_col, wind_glob_max_col+1):
          if glob_row < image_min_row:
            val_row = image_max_row - (image_min_row - glob_row) + 1
          elif glob_row > image_max_row:
            val_row = image_min_row + np.mod(glob_row, image_max_row) - 1
          else:
            val_row = glob_row

          if glob_col < image_min_col:
            val_col = image_max_col - (image_min_col - glob_col) + 1
          elif glob_col > image_max_col:
            val_col = image_min_col + np.mod(glob_col, image_max_col) - 1
          else:
            val_col = glob_col

          window_image[loc_row, loc_col] = image[val_row, val_col]
          loc_col += 1
        loc_row += 1

    return window_image

  def run_texture_analysis(self, feature_postfix, feature_list=None):
    """Run texture analysis"""

    if feature_list is None:
      feature_list = self.feature_list

    if len(feature_list) == 0:
      return

    # Initialize feature images
    for feature_name in feature_list:
      fim = GLCMFeatureImage(np.zeros(self.original_image.shape),
                             feature_name,
                             feature_postfix)
      self.feature_images[fim.name] = fim

    # Create weights related to feature computation
    self.greylevel_prod = np.outer(self.greylevels, self.greylevels).flatten()
    hom_weights_mat = np.zeros([self.num_greylevels, self.num_greylevels])
    con_weights_mat = np.zeros([self.num_greylevels, self.num_greylevels])
    clu_weights_mat = np.zeros([self.num_greylevels, self.num_greylevels])
    for i in range(self.num_greylevels):
      for j in range(self.num_greylevels):
        gi = int(self.greylevels[i])
        gj = int(self.greylevels[j])
        hom_weights_mat[i, j] = 1./(1 + (gi - gj)*(gi - gj))
        con_weights_mat[i, j] = (gi - gj)*(gi - gj)
        clu_weights_mat[i, j] = gi + gj

    self.hom_weights = hom_weights_mat.flatten()
    self.con_weights = con_weights_mat.flatten()
    self.clu_weights = clu_weights_mat.flatten()

    # Check window size relative to glcm feature offset
    window_side_length = 2*self.window_order + 1
    if ((abs(self.delta_x) >= window_side_length) or
            (abs(self.delta_y) >= window_side_length)):
        print('Too large offset in the co-occurance matrix.')
        print('Provided delta_x: {0}'.format(self.delta_x))
        print('Provided delta_y: {0}'.format(self.delta_y))
        print('Window side length: {0}'.format(window_side_length))
        sys.exit(1)

    # Start traversing the pixels of the original image
    total_work = self.num_image_rows * self.num_image_cols
    counter = 0
    win_time = 0.0
    cooc_time = 0.0
    for row_ind in range(self.num_image_rows):
      for col_ind in range(self.num_image_cols):
        counter += 1

        # Cut out window to be analyzed
        temp_win_time = time.time()
        window_image = self.square_image_window(self.original_image,
                                                self.window_order,
                                                row_ind,
                                                col_ind,
                                                self.boundary_treatment)
        win_time += time.time() - temp_win_time

        # Construct grey-level co-occurance matrix
        temp_cooc_time = time.time()

        glcm = self.compute_glcm(window_image, self.delta_x, self.delta_y,
                                 self.greylevels, symmetric=True, normed=True,
                                 isotropic=self.isotropic_glcm)

        flat_glcm = glcm.flatten()
        cooc_time += time.time() - temp_cooc_time

        for feature_name in feature_list:
          self.compute_feature(feature_name, row_ind, col_ind, glcm, flat_glcm)

        if counter % 5000 == 0:
          print('Progress: {0} / {1}  ({2:.2f} %)'.format(counter, total_work,
                                                          counter/total_work*100))

    print('Finished computing feature images')

    print('Computation time summary')
    print('Constructing windows: {0:>7,.2f} seconds.'.format(win_time))
    print('Computing glcm: {0:>7,.2f} seconds.'.format(cooc_time))
    for name, feature_image in self.feature_images.items():
      print('Computing {0:<20}: {1:>7,.2f} seconds.'.format(name, feature_image.computation_time))
